package eu.mhsl.minenet.minigames.instance.room;

import eu.mhsl.minenet.minigames.Resource;
import eu.mhsl.minenet.minigames.instance.MineNetInstance;
import eu.mhsl.minenet.minigames.instance.game.Game;
import eu.mhsl.minenet.minigames.message.Icon;
import eu.mhsl.minenet.minigames.message.type.ChatMessage;
import eu.mhsl.minenet.minigames.util.CommonEventHandles;
import eu.mhsl.minenet.minigames.util.MoveInstance;
import eu.mhsl.minenet.minigames.instance.Spawnable;
import eu.mhsl.minenet.minigames.instance.Dimension;
import eu.mhsl.minenet.minigames.instance.hub.Hub;
import eu.mhsl.minenet.minigames.instance.room.entity.GameSelector;
import net.minestom.server.MinecraftServer;
import net.minestom.server.coordinate.Pos;
import net.minestom.server.entity.GameMode;
import net.minestom.server.entity.Player;
import net.minestom.server.event.player.PlayerBlockBreakEvent;
import net.minestom.server.event.player.PlayerDisconnectEvent;
import net.minestom.server.instance.AnvilLoader;

import java.util.*;
import java.util.logging.Logger;
import java.util.stream.Collectors;

public class Room extends MineNetInstance implements Spawnable {
    private static final Map<Player, Room> rooms = new WeakHashMap<>();
    private static final Logger logger = Logger.getLogger("room");

    public static Room createRoom(Player owner) {
        logger.info("Creating room with owner " + owner.getUsername());
        setRoom(owner, new Room(owner));
        return getRoom(owner);
    }

    public static void deleteRoom(Room room) {
        logger.info("Deleting room with owner " + room.owner.getUsername());
        rooms.values().removeAll(Collections.singleton(room)); // remove(room) would only remove the first one
        room.getAllMembers().forEach(player -> MoveInstance.move(player, Hub.INSTANCE));
        MoveInstance.forceCloseInstance(room);
    }

    public static Room getRoom(Player p) {
        return rooms.get(p);
    }

    public static void setOwnRoom(Player p) {
        setRoom(p, getRoom(p));
    }

    public static void setRoom(Player p, Room room) {
        logger.info("Set room for player " + p.getUsername() + " to room by " + room.owner.getUsername());
        p.clearEffects();
        p.clearTitle();
        p.getInventory().clear();
        p.setGameMode(GameMode.ADVENTURE);
        rooms.put(p, room);
        MoveInstance.move(p, room);
    }

    public static void unsetRoom(Player p) {
        logger.info("Unset room for " + p.getUsername());
        rooms.remove(p);
    }

    public static Set<Room> getAllRooms() {
        return new HashSet<>(rooms.values());
    }

    private Player owner;
    private Room(Player owner) {
        super(Dimension.THE_END.DIMENSION);
        MinecraftServer.getInstanceManager().registerInstance(this);
        setChunkLoader(new AnvilLoader(Resource.LOBBY_MAP.getPath()));

        eventNode().addListener(PlayerBlockBreakEvent.class, CommonEventHandles::cancel);

        eventNode().addListener(PlayerDisconnectEvent.class, playerDisconnectEvent -> unsetRoom(playerDisconnectEvent.getPlayer()));

        setOwner(owner);

        new GameSelector().setInstance(this, new Pos(0.5, 16, 9.5));

    }

    public Player getOwner() {
        return owner;
    }

    private void setOwner(Player newOwner) {
        this.owner = newOwner;

        this.owner.eventNode().addListener(PlayerDisconnectEvent.class, playerDisconnectEvent -> {
            Player p = playerDisconnectEvent.getPlayer();

            if(p != this.owner) return; // return if the current handling player is really the current owner

            getAllMembers().stream()
                    .filter(player -> player != p) // exclude the current leaving owner
                    .findFirst() // find first user meeting the requirement
                    .ifPresentOrElse(
                            this::setOwner, // set the new owner
                            () -> Room.deleteRoom(Room.getRoom(p)) // or else delete the room (no players in the room)
                    );

            Room.unsetRoom(p); // remove the player itself from the room

            new ChatMessage(Icon.ERROR).appendStatic("The room leader left!").send(getAllMembers());
            new ChatMessage(Icon.SCIENCE).appendStatic(this.owner.getUsername()).appendStatic(" is the new Leader!").send(getAllMembers().stream().filter(player -> player != this.owner).collect(Collectors.toSet()));
            new ChatMessage(Icon.SUCCESS).appendStatic("You are now the leader.").send(this.owner);

            logger.info("Room owner changed from " + p.getUsername() + " to " + owner.getUsername());
        });

    }

    public void moveMembersToGame(Game game) {
        logger.info("Move room group with " + this.getAllMembers().size() + " players to Game " + game.getUniqueId());
        this.getAllMembers().forEach(player -> MoveInstance.move(player, game));
    }

    public Set<Player> getAllMembers() {
        return rooms.keySet().stream()
                .filter(player -> getRoom(player) == this)
                .collect(Collectors.toSet());
    }

    @Override
    public Pos getSpawn() {
        return new Pos(0.5, 16, 0.5);
    }
}
